from .functional import revgrad
from torch.nn import Module
from torch import tensor


class RevGrad(Module):
    def __init__(self, alpha=1., *args, **kwargs):
        """
        A gradient reversal layer.

        This layer has no parameters, and simply reverses the gradient
        in the backward pass.
        """
        super().__init__(*args, **kwargs)

        self._alpha = tensor(alpha, requires_grad=False)

    def forward(self, input_):
        return revgrad(input_, self._alpha)



import copy
import pytest
import torch
from pytorch_revgrad import RevGrad


def test_gradients_inverted():
    network = torch.nn.Sequential(torch.nn.Linear(5, 3), torch.nn.Linear(3, 1))
    revnetwork = torch.nn.Sequential(copy.deepcopy(network), RevGrad())

    inp = torch.randn(8, 5)
    outp = torch.randn(8, 1)

    criterion = torch.nn.MSELoss()

    criterion(network(inp), outp).backward()
    criterion(revnetwork(inp), outp).backward()

    assert all(
        (p1.grad == -p2.grad).all()
        for p1, p2 in zip(network.parameters(), revnetwork.parameters())
    )


@pytest.mark.parametrize(
    ("alpha_parameter"), [(0.5), (0.7), (0.1), (30.)],
)
def test_gradients_inverted_alpha(alpha_parameter):
    network = torch.nn.Sequential(torch.nn.Linear(5, 3), torch.nn.Linear(3, 1))
    revnetwork = torch.nn.Sequential(
        copy.deepcopy(network), RevGrad(alpha=alpha_parameter)
    )

    inp = torch.randn(8, 5)
    outp = torch.randn(8, 1)

    criterion = torch.nn.MSELoss()

    criterion(network(inp), outp).backward()
    criterion(revnetwork(inp), outp).backward()

    for p1, p2 in zip(network.parameters(), revnetwork.parameters()):
        assert torch.isclose(p1.grad, -p2.grad/alpha_parameter).all()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pytorch_revgrad']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['numpy', 'torch>=1.0.0']

extras_require = \
{'dev': ['ipython'],
 'test': ['pytest', 'pytest-cov', 'flake8', 'pytest-flake8', 'coveralls']}

setup(name='pytorch_revgrad',
      version='0.2.0',
      description='A pytorch module (and function) to reverse gradients.',
      author='Jan Freyberg',
      author_email='jan.freyberg@gmail.com',
      url='https://github.com/janfreyberg/pytorch-revgrad',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.5',
     )
